#!/usr/bin/ruby
#
# Turing -- Ruby implementation of Captcha
#
# Copyright © 2005 Michal Šafránek <wejn@box.cz>
#
# This file is part of http://turing.rubyforge.org/
#
# See turing.rb in lib/ directory for license terms.
#
require 'pstore'
require 'turing'
require 'turing/image'
require 'digest/sha2'

# == Captcha challenge generator and verifier
# Purpose of this class is to provide abstraction layer (on top of PStore
# and Turing::Image) you can use to build Captcha challenge/response
# mechanism.
#
# Example of use:
#   tc = Turing::Challenge.new(:store => 'store', :outdir => '.')
#   c = tc.generate_challenge
#
#   system("xv", c.file)
#
#   puts "Enter solution:"
#   r = $stdin.gets.chomp
#
#   if tc.valid_answer?(c.id, r)
#       puts "That's right."
#   else
#       puts "I don't think so."
#   end
# In this example records about generated challenges are stored in file
# +store+ which is simple PStore. Images are generated via Turing::Image
# to current directory and then displayed via "xv" image viewer.
class Turing::Challenge # {{{
	# Configure instance using options hash.
	#
	# *Warning*: Keys of this hash must be symbols.
	#
	# Accepted options:
	# * +store+: File to be used as PStore for challenges. Default: <tt>$TMPDIR/turing-challenges.pstore</tt>.
	# * +dictionary+: Filename to be used as dictionary (base for random words). Default: gem's <tt>shared/dictionary</tt> file.
	# * +lifetime+: Lifetime for generated challenge in seconds (to prevent "harvesting").
	# * +outdir+: Outdir for images generated by Turing::Image. Default: <tt>$TMPDIR</tt>.
	#
	# Given hash will be also used to initialize Turing::Image object.
	def initialize(opts = {}) # {{{
		raise ArgumentError, "Opts must be hash!" unless opts.kind_of? Hash

		tmpdir = ENV["TMPDIR"] || '/tmp'
		base = File.join(File.dirname(__FILE__), '..', '..', 'shared')
		@options = {
			:store => File.join(tmpdir, 'turing-challenges.pstore'),
			:dictionary => File.join(base, 'dictionary'),
			:lifetime => 10*60, # 10 minutes
			:outdir => tmpdir,
		}

		@options.merge!(opts)

		begin
			@store = PStore.new(@options[:store])
		rescue
			raise ArgumentError, "Failed to initialize store: #{$!}"
		end

		begin
			File.open(@options[:dictionary]) do |f|
				@dictionary = f.readlines.map! { |x| x.strip }
			end
		rescue
			raise ArgumentError, "Failed to load dictionary: #{$!}"
		end

		begin
			@ti = Turing::Image.new(@options)
		rescue
			raise ArgumentError, "Failed to initialize Turing::Image: #{$!}"
		end
	end # }}}

	# Challenge object we store
	ChallengeObject = Struct.new(:answer, :when) # :nodoc:

	# Generated challenge Struct -- returned from generate_challenge
	GeneratedChallenge = Struct.new(:file, :id)

	# Generate challenge (image containing random word from configured
	# dictionary) and return +GeneratedChallenge+ containing +file+
	# (basename) and +id+ of this challenge.
	#
	# Generation of challenge is retried three times -- to descrease possibility
	# it will fail due to a bug in plugin. But if that happens, we just raise
	# +RuntimeError+.
	def generate_challenge # {{{
		id = nil
		word = nil
		tries = 3
		err = nil
		fname = nil
		
		begin
			id = random_id
			fname = id + ".jpg"
			word = @dictionary[rand(@dictionary.size)]
			@ti.generate(fname, word)
		rescue Object => err
			tries -= 1
			retry if tries > 0
		end
		raise "Failed to generate: #{err}" unless err.nil?

		begin
			@store.transaction do
				@store[id] = ChallengeObject.new(word, Time.now)
			end
		rescue
			raise "Failed to save to store: #{$!}"
		end
		
		GeneratedChallenge.new(fname, id)
	end # }}}

	# Check if +answer+ for challenge with given +id+ is valid.
	#
	# Also removes image file and challenge from the store.
	def valid_answer?(id, answer) # {{{
		ret = false
		begin
			@store.transaction do
				object = @store[id]

				# out if not found
				break if object.nil?

				# remove from store and delete img
				@store.delete(id)
				begin
					n = File.join(@options[:outdir], id + '.jpg')
					File.unlink(n)
				rescue Object
				end

				# true if it's ok
				if object.answer == answer && \
						Time.now < object.when + (@options[:lifetime] || 0)
					ret = true
				end
			end
		rescue
		end
		ret
	end # }}}

	private
	# SHA algorithm "strength" for +random_id+ generation
	SHA_STRENGTH = 256 # 256, 384, 512
	
	# Random device used for +random_id+ generation
	#
	# *Warning*: don't use /dev/random unless you have LOTS of entropy available!
	RANDOM_DEVICE = '/dev/urandom' # '/dev/random'

	# generate random ID using SHA* algo family
	def random_id # {{{
		strength = SHA_STRENGTH
		algo = nil
		begin
			algo = Digest.const_get("SHA" + strength.to_s)
		rescue
			# We got wrong SHA_STRENGTH, fallback to default (256) if possible
			unless strength == 256
				strength = 256
				retry
			else
				raise RuntimeError, "SHA hash algorithm family not available"
			end
		end
		begin
			algo.hexdigest(File.open(RANDOM_DEVICE).sysread(strength / 8 + 1))
		rescue
			# XXX: fall-back, but shouldn't happen
			poor_randomness = [
				Process.pid, rand(65535), Time.now.to_i, Time.now.usec,
				rand(65535)].map { |x| x.to_s }.join(':')
			algo.hexdigest(poor_randomness)
		end
	end # }}}
end # }}}

# test it if invoked directly
if __FILE__ == $0 # {{{
	tc = Turing::Challenge.new(:store => 'store', :outdir => '.')

	c = tc.generate_challenge

	system("xv", c.file)

	puts "Enter solution:"
	r = $stdin.gets.chomp

	if tc.valid_answer?(c.id, r)
		puts "That's right."
	else
		puts "I don't think so."
	end

	exit 0
end # }}}

# vim: set ts=4 sw=4 :
