#!/usr/bin/ruby
#
# Turing -- Ruby implementation of Captcha
#
# Copyright © 2005 Michal Šafránek <wejn@box.cz>
#
# This file is part of http://turing.rubyforge.org/
#
# See turing.rb in lib/ directory for license terms.
#
require 'rubygems'
gem 'gd2'
require 'gd2'
require 'turing'
require 'pathname'

# == Simple obfuscated image generator with plugin design
#
# Generates obfuscated image containing given word by using
# one of its registered plugins.
#
# For it's operation it requires gd2 gem, available from http://gd2.rubyforge.org/.
#
# Example of use:
#   gem 'turing'
#   require 'turing'
#   ti = Turing::Image.new(:width => 280, :height => 115)
#   ti.generate(File.join(Dir.getwd, 'a.jpg'), "randomword")
# In this case we generate image using random plugin containing
# word "randomword". It is saved as <tt>`pwd`/a.jpg</tt>.
#
# Example Rails controller (action):
#   # Could be placed in config/environment.rb
#   gem 'turing'
#   require 'turing'
#   
#   # Could be part of app/controllers/site_controller.rb
#   class SiteController < ApplicationController
#       def image
#           ti = ::Turing::Image.new(:width => 280, :height => 115)
#           fn = get_tmpname
#           ti.generate(fn, rand(1e8).to_s)
#           send_file fn, :type => "image/jpeg", :disposition => "inline"
#       end 
#       
#       def get_tmpname
#           pat = "tmpf-%s-%s-%s"
#           fn = pat % [Process::pid, Time.now.to_f.to_s.tr(".",""), rand(1e8)]
#           File.join(Dir::tmpdir, fn)
#       end 
#       private :get_tmpname
#   end 
#
# === A word about plugins
# All plugins are "registered" by subclassing Turing::Image (which is
# implemented using <tt>self.inherited</tt>). It makes sense to subclass
# Turing::Image because that way you'll also get goodies like #write_string.
#
# Plugins are auto-loaded by require from <tt>lib/turing/image_plugins</tt>
# after Turing::Image is created but you're free to manually load any plugin
# you like.
#
# For inspiration on how to write new plugin visit any of the existing
# plugins in +image_plugins+ dir, minimal template would be:
#  class MyCoolPlugin < Turing::Image
#      def initialize(opts = {})
#          super(opts)
#      end
#  
#      def generate(img, word)
#          write_string(img, 'cour.ttf', GD2::Color[0, 0, 0], word, 48)
#      end
#  end
class Turing::Image # {{{
	# All content generating plugins
	@@plugins = []
	
	# Configure instance using options hash.
	#
	# *Warning*: Keys of this hash must be symbols.
	#
	# Accepted options:
	# * +fontdir+: Directory containing .ttf fonts required by plugins. Default: gem's <tt>shared/fonts</tt> directory.
	# * +bgdir+: Directory containing .jpeg files used as background by plugins. Default: gem's <tt>shared/bgs</tt> directory.
	# * +outdir+: Output directory where to put image in case absolute path wasn't specified.
	# * +width+: Width of the image.
	# * +height+: Height of the image.
	# * +method+: Use specified plugin instead of randomly selected. You must give class that implements +generate+ instance method. Default: nil.
	def initialize(opts = {}) # {{{
		raise ArgumentError, "Opts must be hash!" unless opts.kind_of? Hash
		
		base = File.join(File.dirname(__FILE__), '..', '..', 'shared')
		@options = {
			:fontdir => File.join(base, 'fonts'),
			:bgdir => File.join(base, 'bgs'),
			:outdir => ENV["TMPDIR"] || '/tmp',
			:width => 280,
			:height => 115,
		}

		@options.merge!(opts)
	end # }}}

	# Generate image into +outname+ containing +word+ (using +method+).
	#
	# *Warning*: If you pass absolute filename as +outname+, +outdir+ will have no effect.
	#
	# *Warning*: There's no way to reset +method+ to random if it was specified upon instance creation.
	def generate(outname, word, method = nil) # {{{
		# select appropriate output plugin # {{{
		m = method || @options[:method]
		if m.nil?
			if @@plugins.empty?
				raise RuntimeError, "no generators plugins available!"
			end
			m = @@plugins[rand(@@plugins.size)]
		end
		unless m.instance_methods.include?("generate")
			raise ArgumentError, "plugin #{m} doesn't have generate method"
		end
		# }}}

		# prepend outname with outdir, if no absolute path given
		unless Pathname.new(outname).absolute?
			outname = File.join(@options[:outdir], outname)
		end
		
		img = GD2::Image.new(@options[:width], @options[:height])

		img.draw do |canvas|
			canvas.color = GD2::Color[255, 255, 255]
			canvas.rectangle(0, 0, img.width - 1, img.height - 1, true)
		end

		m.new(@options).generate(img, word)
		
		img.draw do |canvas|
			canvas.color = GD2::Color[0, 0, 0]
			canvas.rectangle(0, 0, img.width - 1, img.height - 1)
		end

		begin
			File.open(outname, 'w') { |f| f.write(img.jpeg(90)) }
		rescue
			raise "Unable to write challenge: #{$!}"
		end

		true
	end # }}}

	# Private methods
	private

	# Write +string+ to +img+ using color +fg+ and +font+
	# (with size +req_size+, if possible) at random coordinates
	# and using random angle.
	#
	# Method checks bounding box so the string is guaranteed to stay
	# within image's dimensions.
	#
	# May raise RuntimeError if it's completely impossible to find suitable
	# fontsize for given dimensions.
	def write_string(img, font, fg, string, req_size = nil) # {{{ # :doc:
		# prepend fontname with fontdir, unless absolute path given
		unless Pathname.new(font).absolute?
			font = File.join(@options[:fontdir], font)
		end
		sizes = (16..42).to_a.reverse
		turbulence = 5 # x%
		mult = 0.85 * ((rand(turbulence*2 + 1) - turbulence) / 100.0 + 1.0)

		# select angle
		angle = -5 + rand(11)
		
		# font size determination ...
		chosen = nil # {{{
		sizes.unshift(req_size) unless req_size.nil?
		sizes.each do |size|
			bounds = nil
			begin
				bounds = GD2::Font::TrueType.new(font, size).
					bounding_rectangle(string, angle.degrees)
			rescue
				raise "Unable to detect bounding box: #{$!}"
			end

			minx, maxx = bounds.values.map { |x| x[0] }.sort.values_at(0, -1)
			miny, maxy = bounds.values.map { |x| x[1] }.sort.values_at(0, -1)

			bb_width = maxx - minx
			bb_height = maxy - miny

			if img.width * mult > bb_width && img.height * mult > bb_height
				chosen = {
					:size => size,
					:width => bb_width,
					:height => bb_height,
				}
				break
			end
		end # }}}

		raise "Unable to select size" if chosen.nil?
		
		x_base, y_base = 1, img.height / 2

		x_offset = rand(((img.width - chosen[:width])*mult).to_i)
		y_offset = rand((((img.height - chosen[:height]) / 2)*mult).to_i)

		x = x_base + x_offset
		y = y_base + y_offset

		img.draw do |canvas|
			canvas.move_to(x, y)
			canvas.color = fg
			canvas.font = GD2::Font::TrueType.new(font, chosen[:size])
			canvas.text(string, angle.degrees)
		end

		img
	end # }}}

	# Record all plugins available
	# (subclasses which implement class method generate)
	def self.inherited(other) # {{{
		(@@plugins ||= []) << other
	end # }}}
end # }}}

# Load all image plugins ...
# {{{
plugin_dir = File.join(File.dirname(__FILE__), 'image_plugins')
if FileTest.directory?(plugin_dir)
	Dir[File.join(plugin_dir, '*')].each do |f|
		bn = File.basename(f)
		begin
			require 'turing/image_plugins/' + bn
		rescue Object
			raise RuntimeError, "Exception while loading plugin: #{bn}: #{$!}"
		end
	end
end
# }}}

# Test it out, when executed directly
if __FILE__ == $0 # {{{
	ti = Turing::Image.new(:width => 280, :height => 115)

	# debug loop
	begin # {{{
		loop do
			dict_fn = File.join(File.dirname(__FILE__), '..', '..', 'shared',
				'dictionary')
			dict = File.open(dict_fn).readlines.map! { |x| x.chomp }
			ti.generate(File.join(Dir.getwd, 'a.jpg'),
				dict[rand(dict.size)])
			pid = fork do
				exec("xv", "a.jpg")
			end
			Process.detach(pid)
			sleep 2
			Process.kill(9, pid)
		end
	rescue Interrupt
	ensure
		File.unlink('a.jpg') rescue nil
	end # }}}
end # }}}

# vim: set ts=4 sw=4 :
